/*
 * This file: xrdbm.c 
 *
 * This file contains functions that allow you to merge some resources 
 * specified in a buffer with those already on some display.  The ONLY
 * entry point into this file is the function "merge_resources".  Call
 * that function and when it returns the resources will be merged.
 *
 * NOTE: if you use "XtDatabase" in your program, you will need to close
 * the server connection, then re-open it using XtOpenDisplay, since this
 * is the only way to read in the new database.
 *
 * The functions in this file were highly derived from the X11R4 client xrdb 
 * source code which was originally written by Jim Gettys.  I didn't clean
 * up the source code much at all for lack of time.  Therefore, all bad
 * code came from Jim, and all the brilliant stuff came from me.  :-)
 *
 * Author: Brian Wilson (brianw@apple.com) on 4/25/92
 */

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xos.h>

#define INIT_BUFFER_SIZE 10000
#define INIT_ENTRY_SIZE 500

void AppendToBuffer();
void GetEntries();

typedef struct _Entry {
    char *tag, *value;
    int lineno;
    Bool usable;
} Entry;

typedef struct _Buffer {
    char *buff;
    int  room, used;
} Buffer;

typedef struct _Entries {
    Entry *entry;
    int   room, used;
} Entries;


/*
 * merge_resources - This function is the ONLY entry point into this file's 
 *                   functions.  This function merges the resources that are 
 *                   specified in the buffer onto the indicated display.  
 *                   "Merge" means just like "xrdb -merge".  A typical
 *                   call might look like:
 *                   "merge_resources(dpy, "foo*bar: 5\nfoo*bar2: red\n");"
 */
merge_resources(disp, buff)
Display *disp;
char *buff;
{
    char *xdefs;
    Buffer buffer;
    Entries newDB, oldDB;
    char *saveBuff;

    buffer.buff = (char *) malloc(INIT_BUFFER_SIZE*sizeof(char));
    strncpy(buffer.buff, buff, INIT_BUFFER_SIZE - 1);
    buffer.used = strlen(buffer.buff);
    buffer.room = INIT_BUFFER_SIZE - buffer.used - 1;

    newDB.room = INIT_ENTRY_SIZE;
    newDB.used = 0;
    newDB.entry = (Entry *)malloc(INIT_ENTRY_SIZE*sizeof(Entry));

    oldDB.room = INIT_ENTRY_SIZE;
    oldDB.used = 0;
    oldDB.entry = (Entry *)malloc(INIT_ENTRY_SIZE*sizeof(Entry));

    GetEntries(&newDB, &buffer);

    xdefs = XResourceManagerString(disp);

    if (xdefs) 
    {
        saveBuff = buffer.buff;
        buffer.used = strlen(xdefs);
        buffer.buff = xdefs;
	GetEntries(&oldDB, &buffer);
	buffer.buff = saveBuff;
    } 
    else
        oldDB.used = 0;

    buffer.used = 0;
    MergeEntries(&buffer, newDB, oldDB);

    XChangeProperty(disp, RootWindow(disp, 0),
		    XA_RESOURCE_MANAGER, XA_STRING, 8, PropModeReplace,
    		    (unsigned char *) buffer.buff, buffer.used);
    /*
     * THIS SHOULD WORK, but does not.
     *
    XChangeProperty(disp, RootWindow(disp, DefaultScreen(disp)),
		    XA_RESOURCE_MANAGER, XA_STRING, 8, PropModeReplace,
    		    (unsigned char *) buffer.buff, buffer.used);
    */

    free(buffer.buff);
}


void AddEntry(e, entry)
register Entries *e;
Entry entry;
{
    register int n;

    for (n = 0; n < e->used; n++) 
    {
	if (strcmp(e->entry[n].tag, entry.tag) == 0)  /* overwrite old entry */
	{ 
	    e->entry[n] = entry;
	    return ;  /* ok to leave, now there's only one of each tag in db */
	}
    }

    if (e->used == e->room) 
    {
	e->entry = (Entry *)realloc(e->entry, 2*e->room*(sizeof(Entry)));
	e->room *= 2;
    }
    entry.usable = True;
    e->entry[e->used++] = entry;
}


int CompareEntries(e1, e2)
Entry *e1, *e2;
{
    return strcmp(e1->tag, e2->tag);
}

void AppendEntryToBuffer(buffer, entry)
register Buffer *buffer;
Entry entry;
{
    AppendToBuffer(buffer, entry.tag, strlen(entry.tag));
    AppendToBuffer(buffer, ":\t", 2);
    AppendToBuffer(buffer, entry.value, strlen(entry.value));
    AppendToBuffer(buffer, "\n", 1);
}

char *FindFirst(string, dest)
register char *string;
register char dest;
{
    for (;;) {
	if (*string == '\0')
	    return NULL;
	if (*string == '\\') {
	    if (*++string == '\0')
		return NULL;
	}
	else if (*string == dest)
	    return string;
	string++;
    }
}


void GetEntries(entries, buff)
register Entries *entries;
Buffer *buff;
{
    register char *line, *colon, *temp, *str;
    Entry entry;
    register int length;
    int dosort = 1;	/* merge */
    int lineno = 0;

    str = buff->buff;
    if (!str) return;
    for (; str < buff->buff + buff->used; str = line + 1) {
	line = FindFirst(str, '\n');
	lineno++;
	if (line == NULL)
	    break; 
	if (str[0] == '!')
	    continue;
	if (str[0] == '\n')
	    continue;
	if (str[0] == '#') {
	    int dummy;
	    if (sscanf (str, "# %d", &dummy) == 1) lineno = dummy - 1;
	    continue;
	}
	for (temp = str; 
	     *temp && *temp != '\n' && isascii(*temp) && isspace(*temp); 
	     temp++) ;
	if (!*temp || *temp == '\n') continue;

	colon = FindFirst(str, ':');
	if (colon == NULL)
	    break;
	if (colon > line) {
	    line[0] = '\0';
	    fprintf(stderr, "Colon missing on line %d, ignoring entry \"%s\"\n",
		    lineno, str);
	    continue;
	}

	temp = (char *)malloc((length = colon - str) + 1);
	strncpy(temp, str, length);
	temp[length] = '\0';
	while (temp[0] == ' ' || temp[0] == '\t') {
	    temp++; length--;
	}
	while (temp[length-1] == ' ' || temp[length-1] == '\t')
	    temp[--length] = '\0';
	entry.tag = temp;

	temp = (char *)malloc((length = line - colon - 1) + 1);
	strncpy(temp, colon + 1, length);
	temp[length] = '\0';
	while (temp[0] == ' ' || temp[0] == '\t') {
	    temp++; length--;
	}
	entry.value = temp;
	entry.lineno = lineno;

	AddEntry(entries, entry);
    }
    if (dosort && (entries->used > 0))
      qsort(entries->entry, entries->used, sizeof(Entry), CompareEntries);
}


int MergeEntries(buffer, new, old)
Entries new, old;
Buffer *buffer;
{
    int n, o, cmp;

    buffer->used = 0;
    n = o = 0;
    while ((n < new.used) && (o < old.used)) {
	cmp = strcmp(new.entry[n].tag, old.entry[o].tag);
	if (cmp > 0)
	    AppendEntryToBuffer(buffer, old.entry[o++]);
	else {
	    AppendEntryToBuffer(buffer, new.entry[n++]);
	    if (cmp == 0)
		o++;
	}
    }
    while (n < new.used)
	AppendEntryToBuffer(buffer, new.entry[n++]);
    while (o < old.used)
	AppendEntryToBuffer(buffer, old.entry[o++]);
    AppendToBuffer(buffer, "\0", 1);
}


void AppendToBuffer(b, str, len)
register Buffer *b;
char *str;
register int len;
{
    while (b->used + len > b->room) {
	b->buff = (char *)realloc(b->buff, 2*b->room*(sizeof(char)));
	b->room *= 2;
    }
    strncpy(b->buff + b->used, str, len);
    b->used += len;
}

